@@ -10,13 +10,13 @@ from distutils import version
 
 init()
 
-__version__ = '0.4.4'
+__version__ = '0.4.5'
 
 try:
-    user_python_verison = sys.version_info
-    if not (user_python_verison.major >= MIN_PYTHON_MAJOR and user_python_verison.minor >= MIN_PYTHON_MINOR):
+    user_python_version = sys.version_info
+    if not (user_python_version.major >= MIN_PYTHON_MAJOR and user_python_version.minor >= MIN_PYTHON_MINOR):
         print(f"{Style.BRIGHT}{Fore.RED}\tIt is highly advised that you use Python {MIN_PYTHON_MAJOR}.{MIN_PYTHON_MINOR} or greater...\n"
-              f"\tSome features may not work as intended on your version ({user_python_verison.major}.{user_python_verison.minor}.{user_python_verison.micro})...{Style.RESET_ALL}")
+              f"\tSome features may not work as intended on your version ({user_python_version.major}.{user_python_version.minor}.{user_python_version.micro})...{Style.RESET_ALL}")
         print(f"\tPlease decide if you want to proceed with running as is or not...\n")
         wrong_version_warning_input = input(f"\tReply with: {Back.CYAN}{Fore.BLACK}Y/N{Style.RESET_ALL} (Anything other than {Back.CYAN}{Fore.BLACK}Y{Style.RESET_ALL} will exit the code) : ")
         if wrong_version_warning_input.upper()[0] != "Y":
