@@ -6,9 +6,11 @@ from threading import Thread
 from datetime import datetime
 from dateutil.relativedelta import relativedelta
 import json
+import time
 
 from colorama import Style, Fore
 from builtins import TimeoutError, ConnectionResetError
+from copy import deepcopy
 
 DEFAULT_HISTORIC_DATA_ID = 1000
 DEFAULT_MARKET_DATA_ID = 2000
@@ -39,7 +41,61 @@ class LiveClient(WebSocketApp):
         self.valid_until = ''
         self.contract_mapping = {}
         self.subscription_type = ''
+        self.confirm_heartbeats = 1
+        self.store_last_n_heartbeats = self.confirm_heartbeats + 7
+        self.heartbeat_interval = 5
+        self.heartbeat_buffer = 0.9
+        time_of_creation = datetime.now()
+        self.last_n_heartbeat = [time_of_creation - relativedelta(seconds=i * self.heartbeat_interval) for i in range(-self.store_last_n_heartbeats, 0)]
+        # print("RECREATED")
         self.parent_app = parent_app
+        self.disconnect_flag = False
+        self.heartbeat_check_thread = Thread(target=self.check_heartbeat)
+        if self.parent_app.live_port == 8086:
+            self.heartbeat_check_thread.start()
+
+    def check_heartbeat(self):
+        while True:
+            time.sleep(self.heartbeat_interval)
+            if self.disconnect_flag:
+                print(f"{Fore.GREEN}Removing hand from the pulse...{Style.RESET_ALL}")
+                break
+            base_heartbeat = self.last_n_heartbeat[-self.confirm_heartbeats]
+            check_time = datetime.now()
+            time_diff = check_time - base_heartbeat
+            if time_diff.total_seconds() > ((self.heartbeat_interval + self.heartbeat_buffer) * self.confirm_heartbeats):
+                print(f"{Style.BRIGHT}{Fore.RED}Failed Heartbeat @ {check_time} because of last at {base_heartbeat}{Style.RESET_ALL}")
+                print(f"{Style.BRIGHT}{Fore.RED}Attempting reconnect @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
+                restart_successful = self.reconnect()
+                if restart_successful:
+                    print(f"{Style.BRIGHT}{Fore.RED}Successful restart @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
+                    time.sleep((self.heartbeat_interval + self.heartbeat_buffer))
+                    recover_start, recover_end = self.get_largest_diff(self.last_n_heartbeat)
+                    # print(f"\t\t\t{len(self.last_n_heartbeat)} - {self.last_n_heartbeat}")
+                    self.recover_from_time_missed(recover_start, recover_end)
+                else:
+                    print(f"{Style.BRIGHT}{Fore.RED}Failed restart @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
+            # else:
+            #     pass
+            #     print(f"{Style.BRIGHT}{Fore.GREEN}Successful Heartbeat @ {check_time} because of last at {base_heartbeat}{Style.RESET_ALL}")
+
+    @staticmethod
+    def get_largest_diff(time_series):
+        big_li = deepcopy(time_series[1:])
+        small_li = deepcopy(time_series[:-1])
+        diffs = [i[0]-i[1] for i in zip(big_li, small_li)]
+        start_gap_index = max(range(len(diffs)), key=lambda i: diffs[i])
+        return time_series[start_gap_index], time_series[start_gap_index+1]
+
+    def recover_from_time_missed(self, start_time, end_time):
+        print(f"{Style.BRIGHT}{Fore.YELLOW}Initiating recovery from {Fore.GREEN}{start_time}{Fore.YELLOW} till {Fore.GREEN}{end_time}{Fore.YELLOW} which are last green heartbeats from server...{Style.RESET_ALL}")
+
+    def reconnect(self):
+        self.close()
+        t = Thread(target=self.run_forever)
+        t.start()
+        time.sleep(3)  # TODO: 3 second restart time hardcoded
+        return t.is_alive()
 
     def on_msg_func(self, message):
         msg = json.loads(message)
@@ -59,7 +115,9 @@ class LiveClient(WebSocketApp):
     def handle_message_data(self, msg):
         # print(f"Truedata has this message for you -> {msg['message']}")
         if msg['success']:
-            if msg['message'] == 'TrueData Real Time Data Service':  # Connection success message
+            if msg['message'] == 'HeartBeat':
+                self.handle_heartbeat(msg['timestamp'])
+            elif msg['message'] == 'TrueData Real Time Data Service':  # Connection success message
                 print(f"You have subscribed for {msg['maxsymbols']} symbols across {msg['segments']} until {msg['validity']} with type of stream as {msg['subscription']}...")
                 self.subscription_type = msg['subscription']
             elif msg['message'] == 'symbols added':
@@ -67,7 +125,12 @@ class LiveClient(WebSocketApp):
             elif msg['message'] == 'symbols removed':
                 pass
         else:
-            print(f"{Style.BRIGHT}{Fore.RED}The request encountered an error - {msg['message']}{Style.BRIGHT}{Fore.RED}")
+            print(f"{Style.BRIGHT}{Fore.RED}The request encountered an error - {msg['message']}{Style.RESET_ALL}")
+
+    def handle_heartbeat(self, server_timestamp):
+        timestamp = datetime.strptime(server_timestamp + ((26 - len(server_timestamp)) * "0"), "%Y-%m-%d %H:%M:%S.%f")
+        self.last_n_heartbeat = self.last_n_heartbeat[1:]
+        self.last_n_heartbeat.append(timestamp)
 
     def add_contract_details(self, contracts_list):
         for contract in contracts_list:
@@ -85,7 +148,7 @@ class LiveClient(WebSocketApp):
                 self.parent_app.touchline_data[ticker_id].oi = int(contract_details[8])
                 self.parent_app.touchline_data[ticker_id].prev_oi = int(contract_details[9])
                 self.parent_app.touchline_data[ticker_id].turnover = float(contract_details[10])
-                self.parent_app.touchline_data[ticker_id].tick_type = 0
+                self.parent_app.live_data[ticker_id].tick_type = 0
                 self.parent_app.live_data[ticker_id].populate_using_touchline(self.parent_app.touchline_data[ticker_id])
 
     def handle_trade_data(self, trade_tick):
@@ -107,8 +170,10 @@ class LiveClient(WebSocketApp):
                         self.parent_app.live_data[ticker_id].day_high = self.parent_app.touchline_data[ticker_id].high = ltp
                     elif special_tag == 'L':
                         self.parent_app.live_data[ticker_id].day_low = self.parent_app.touchline_data[ticker_id].low = ltp
+                    elif special_tag == 'O':
+                        self.parent_app.live_data[ticker_id].day_open = self.parent_app.touchline_data[ticker_id].open = ltp
                 self.parent_app.live_data[ticker_id].tick_seq = int(trade_tick[8])
-                self.parent_app.touchline_data[ticker_id].tick_type = 1
+                self.parent_app.live_data[ticker_id].tick_type = 1
                 # Calculating addn data
                 self.parent_app.live_data[ticker_id].calculate_additional_data()
                 try:
@@ -117,14 +182,19 @@ class LiveClient(WebSocketApp):
                     self.parent_app.live_data[ticker_id].best_ask_price = float(trade_tick[11])
                     self.parent_app.live_data[ticker_id].best_ask_qty = int(trade_tick[12])
                 except (IndexError, ValueError, TypeError):
-                    del self.parent_app.live_data[ticker_id].best_bid_price
-                    del self.parent_app.live_data[ticker_id].best_bid_qty
-                    del self.parent_app.live_data[ticker_id].best_ask_price
-                    del self.parent_app.live_data[ticker_id].best_ask_qty
+                    try:
+                        del self.parent_app.live_data[ticker_id].best_bid_price
+                        del self.parent_app.live_data[ticker_id].best_bid_qty
+                        del self.parent_app.live_data[ticker_id].best_ask_price
+                        del self.parent_app.live_data[ticker_id].best_ask_qty
+                    except AttributeError:
+                        pass
+                except Exception as e:
+                    print(e)
         except KeyError:
             pass
         except Exception as e:
-            print(f'{Style.BRIGHT}{Fore.RED}Encountered with tick feed - {e}{Style.RESET_ALL}')
+            print(f'{Style.BRIGHT}{Fore.RED}Encountered with tick feed - {type(e)}{Style.RESET_ALL}')
 
     def handle_bid_ask_data(self, bidask_tick):
         try:
@@ -136,7 +206,7 @@ class LiveClient(WebSocketApp):
                 self.parent_app.live_data[ticker_id].best_bid_qty = int(bidask_tick[3])
                 self.parent_app.live_data[ticker_id].best_ask_price = float(bidask_tick[4])
                 self.parent_app.live_data[ticker_id].best_ask_qty = int(bidask_tick[5])
-                self.parent_app.touchline_data[ticker_id].tick_type = 2
+                self.parent_app.live_data[ticker_id].tick_type = 2
         except KeyError:
             pass
         except Exception as e:
@@ -169,7 +239,6 @@ class LiveClient(WebSocketApp):
 
     def ws_error(self, error):
         if any(isinstance(error, conn_error) for conn_error in [ConnectionResetError, TimeoutError]):
-            print('Alright...')
             print(error)
 
     def ws_open(self):
@@ -227,6 +296,18 @@ class HistoricalWebsocket:
                 bar_size = bar_size[:-1]
             return self.get_hist_bar_data(contract, end_time, start_time, bar_size)
 
+    def get_hist_tick_data(self, contract, end_time, start_time):
+        self.hist_socket.send(f'{{"method": "gethistory", "interval": "tick", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
+        raw_hist_data = self.hist_socket.recv()
+        json_response = json.loads(raw_hist_data)
+        if json_response['success']:
+            hist_data = json_response['data']
+            hist_data = self.hist_tick_data_to_dict_list(hist_data, contract)
+            return hist_data
+        else:
+            print(f'{Style.BRIGHT}{Fore.RED}ERROR: Your request failed with error "{json_response["message"]}"{Style.RESET_ALL}')
+            return []
+
     def get_hist_bar_data(self, contract, end_time, start_time, bar_size):
         self.hist_socket.send(f'{{"method": "gethistory", "interval": "{bar_size}", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
         raw_hist_data = self.hist_socket.recv()
@@ -239,26 +320,15 @@ class HistoricalWebsocket:
             print(f'{Style.BRIGHT}{Fore.RED}ERROR: Your request failed with error "{json_response["message"]}"{Style.RESET_ALL}')
             return []
 
-    @staticmethod
-    def hist_bar_data_to_dict_list(hist_data):
-        data_list = []
-        for j in hist_data:
-            data_list.append({'time': datetime.strptime(j[0], '%Y-%m-%dT%H:%M:%S'),
-                              'o': float(j[1]),
-                              'h': float(j[2]),
-                              'l': float(j[3]),
-                              'c': float(j[4]),
-                              'v': int(j[5]),
-                              'oi': int(j[6])})
-        return data_list
-
-    def get_hist_tick_data(self, contract, end_time, start_time):
-        self.hist_socket.send(f'{{"method": "gethistory", "interval": "tick", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
+    def get_hist_eod_data(self, contract, end_time, start_time):
+        start_time = start_time.split('T')[0]
+        end_time = end_time.split('T')[0]
+        self.hist_socket.send(f'{{"method": "gethistory", "interval": "EOD", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
         raw_hist_data = self.hist_socket.recv()
         json_response = json.loads(raw_hist_data)
         if json_response['success']:
             hist_data = json_response['data']
-            hist_data = self.hist_tick_data_to_dict_list(hist_data, contract)
+            hist_data = self.hist_eod_data_to_dict_list(hist_data, contract)
             return hist_data
         else:
             print(f'{Style.BRIGHT}{Fore.RED}ERROR: Your request failed with error "{json_response["message"]}"{Style.RESET_ALL}')
@@ -294,19 +364,18 @@ class HistoricalWebsocket:
             count = count + 1
         return data_list
 
-    def get_hist_eod_data(self, contract, end_time, start_time):
-        start_time = start_time.split('T')[0]
-        end_time = end_time.split('T')[0]
-        self.hist_socket.send(f'{{"method": "gethistory", "interval": "EOD", "symbol": "{contract}", "from": "{start_time}", "to": "{end_time}"}}')
-        raw_hist_data = self.hist_socket.recv()
-        json_response = json.loads(raw_hist_data)
-        if json_response['success']:
-            hist_data = json_response['data']
-            hist_data = self.hist_eod_data_to_dict_list(hist_data, contract)
-            return hist_data
-        else:
-            print(f'{Style.BRIGHT}{Fore.RED}ERROR: Your request failed with error "{json_response["message"]}"{Style.RESET_ALL}')
-            return []
+    @staticmethod
+    def hist_bar_data_to_dict_list(hist_data):
+        data_list = []
+        for j in hist_data:
+            data_list.append({'time': datetime.strptime(j[0], '%Y-%m-%dT%H:%M:%S'),
+                              'o': float(j[1]),
+                              'h': float(j[2]),
+                              'l': float(j[3]),
+                              'c': float(j[4]),
+                              'v': int(j[5]),
+                              'oi': int(j[6])})
+        return data_list
 
     @staticmethod
     def hist_eod_data_to_dict_list(hist_data, contract):  # No need for symbol other than printing
@@ -357,6 +426,7 @@ class TD:
         self.live_websocket.run_forever()
 
     def disconnect(self):
+        self.live_websocket.disconnect_flag = True
         self.live_websocket.close()
         print(f"{Fore.GREEN}Disconnected LIVE TrueData...{Style.RESET_ALL}")
         if self.connect_historical:
